# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import pkg_resources
from unittest import skipUnless
from django.test import TestCase
from tldap.test import slapd

from karaage.middleware.threadlocals import reset
from karaage.tests.initial_ldap_data import test_ldif
from karaage.datastores import _DATASTORES
from karaage.datastores.ldap import DataStore


def skip_if_missing_requirements(*requirements):
    try:
        pkg_resources.require(*requirements)
        msg = ''
    except pkg_resources.DistributionNotFound:
        msg = 'Missing one or more requirements (%s)' % '|'.join(requirements)
    return skipUnless(msg == '', msg)


class IntegrationTestCase(TestCase):
    LDAP_CONFIG = {
        'DESCRIPTION': 'LDAP datastore',
        'ENGINE': 'karaage.datastores.ldap.AccountDataStore',
        'LDAP': 'default',
        'ACCOUNT': 'karaage.datastores.ldap_schemas.openldap_account',
        'GROUP': 'karaage.datastores.ldap_schemas.openldap_account_group',
        'PRIMARY_GROUP': "institute",
        'DEFAULT_PRIMARY_GROUP': "dummy",
        'HOME_DIRECTORY': "/vpac/%(default_project)s/%(uid)s",
        'LOCKED_SHELL': "/usr/local/sbin/locked",
        'LDAP_ACCOUNT_BASE': 'ou=Account,dc=python-ldap,dc=org',
        'LDAP_GROUP_BASE': 'ou=Group,dc=python-ldap,dc=org',
    }

    def setUp(self):
        super(IntegrationTestCase, self).setUp()
        server = slapd.Slapd()
        server.set_port(38911)
        server.start()

        self.addCleanup(self.cleanup)

        server.ldapadd("\n".join(test_ldif) + "\n")
        self.__ldap_server = server
        self._ldap_datastore = DataStore(self.LDAP_CONFIG)
        _DATASTORES.clear()
        _DATASTORES.append(self._ldap_datastore)

    def cleanup(self):
        self.__ldap_server.stop()
        reset()
        _DATASTORES.clear()



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import six
import django_tables2 as tables

from django.db.models import Q
from django.shortcuts import get_object_or_404, render
from django.http import HttpResponseRedirect, HttpResponseForbidden
from django.core.urlresolvers import reverse
from django.forms.models import inlineformset_factory

from karaage.common import is_admin
from karaage.common.decorators import admin_required, login_required
import karaage.common as util
from karaage.institutes.tables import InstituteTable, InstituteFilter
from karaage.institutes.models import Institute, InstituteDelegate
from karaage.institutes.forms import InstituteForm, DelegateForm
from karaage.people.models import Person
from karaage.people.tables import PersonTable
from karaage.projects.tables import ProjectTable


@login_required
def profile_institutes(request):
    config = tables.RequestConfig(request, paginate={"per_page": 5})

    person = request.user

    my_institute_list = Institute.objects.filter(
        Q(pk=person.institute_id) | Q(group__members=person))
    my_institute_list = my_institute_list.select_related()
    my_institute_list = InstituteTable(my_institute_list, prefix="mine")
    config.configure(my_institute_list)

    delegate_institute_list = person.delegate_for.all()
    delegate_institute_list = delegate_institute_list.select_related()
    delegate_institute_list = InstituteTable(
        delegate_institute_list, prefix="delegate")
    config.configure(delegate_institute_list)

    return render(
        template_name='karaage/institutes/profile_institutes.html',
        context={
            'person': person,
            'my_institute_list': my_institute_list,
            'delegate_institute_list': delegate_institute_list},
        request=request)


@login_required
def institute_detail(request, institute_id):
    config = tables.RequestConfig(request, paginate={"per_page": 5})

    institute = get_object_or_404(Institute, pk=institute_id)
    if not institute.can_view(request):
        return HttpResponseForbidden(
            '<h1>Access Denied</h1>'
            '<p>You do not have permission to view details'
            'about this institute.</p>')

    project_list = institute.project_set.select_related()
    project_list = ProjectTable(project_list, prefix="project-")
    config.configure(project_list)

    person_list = Person.objects.filter(
        Q(institute__pk=institute_id) | Q(groups__institute=institute_id))
    person_list = person_list.select_related()
    person_list = PersonTable(person_list, prefix="person-")
    config.configure(person_list)

    return render(
        template_name='karaage/institutes/institute_detail.html',
        context=locals(),
        request=request)


@admin_required
def institute_verbose(request, institute_id):
    institute = get_object_or_404(Institute, pk=institute_id)

    from karaage.datastores import get_institute_details
    institute_details = get_institute_details(institute)

    return render(
        template_name='karaage/institutes/institute_verbose.html',
        context=locals(),
        request=request)


@login_required
def institute_list(request):

    queryset = Institute.objects.all()
    if not is_admin(request):
        queryset = institute_list.filter(
            is_active=True, delegates=request.user)

    q_filter = InstituteFilter(request.GET, queryset=queryset)
    table = InstituteTable(q_filter.qs)
    tables.RequestConfig(request).configure(table)

    spec = []
    for name, value in six.iteritems(q_filter.form.cleaned_data):
        if value is not None and value != "":
            name = name.replace('_', ' ').capitalize()
            spec.append((name, value))

    return render(
        template_name='karaage/institutes/institute_list.html',
        context={
            'table': table,
            'filter': q_filter,
            'spec': spec,
            'title': "Institute list",
        },
        request=request)


@admin_required
def add_edit_institute(request, institute_id=None):

    if institute_id:
        institute = get_object_or_404(Institute, pk=institute_id)
    else:
        institute = None

    delegate_formset_class = inlineformset_factory(
        Institute, InstituteDelegate, form=DelegateForm, extra=3,
        min_num=1, validate_min=True)

    if request.method == 'POST':
        form = InstituteForm(request.POST, instance=institute)

        delegate_formset = delegate_formset_class(
            request.POST, instance=institute)

        if form.is_valid():
            institute = form.save()

            if delegate_formset.is_valid():
                delegate_formset.save()
                return HttpResponseRedirect(institute.get_absolute_url())
    else:
        form = InstituteForm(instance=institute)
        delegate_formset = delegate_formset_class(instance=institute)

    media = form.media
    for dform in delegate_formset.forms:
        media = media + dform.media

    return render(
        template_name='karaage/institutes/institute_form.html',
        context={
            'institute': institute, 'form': form,
            'media': media, 'delegate_formset': delegate_formset},
        request=request)


@admin_required
def institute_logs(request, institute_id):
    obj = get_object_or_404(Institute, pk=institute_id)
    breadcrumbs = [
        ("Institutes", reverse("kg_institute_list")),
        (six.text_type(obj), reverse("kg_institute_detail", args=[obj.pk]))
    ]
    return util.log_list(request, breadcrumbs, obj)


@admin_required
def add_comment(request, institute_id):
    obj = get_object_or_404(Institute, pk=institute_id)
    breadcrumbs = [
        ("Institutes", reverse("kg_institute_list")),
        (six.text_type(obj), reverse("kg_institute_detail", args=[obj.pk]))
    ]
    return util.add_comment(request, breadcrumbs, obj)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django_tables2 as tables
from django_tables2.utils import A
import django_filters

from django.utils.safestring import mark_safe

from .models import Institute
from karaage.people.tables import PeopleColumn


class ActiveFilter(django_filters.ChoiceFilter):

    def __init__(self, *args, **kwargs):
        choices = [
            ('', 'Unknown'),
            ('deleted', 'Deleted'),
            ('yes', 'Yes'),
        ]

        super(ActiveFilter, self).__init__(*args, choices=choices, **kwargs)

    def filter(self, qs, value):
        if value == "deleted":
            qs = qs.filter(is_active=False)
        elif value == "yes":
            qs = qs.filter(is_active=True)

        return qs


class InstituteFilter(django_filters.FilterSet):
    active = ActiveFilter()
    name = django_filters.CharFilter(lookup_expr="icontains")

    class Meta:
        model = Institute
        fields = ('name', 'active')


class InstituteTable(tables.Table):
    is_active = tables.Column(order_by='-is_active', verbose_name="active")
    name = tables.LinkColumn('kg_institute_detail', args=[A('pk')])
    delegates = PeopleColumn(orderable=False)

    def render_is_active(self, record):
        if not record.is_active:
            html = '<span class="no">Deleted</span>'
        else:
            html = '<span class="yes">Yes</span>'
        return mark_safe(html)

    class Meta:
        model = Institute
        fields = ('is_active', 'name')
        empty_text = "No items"



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

""" LDAP datastore. """

from karaage.datastores import base

import importlib
import logging
import karaage.common.trace as trace
logger = logging.getLogger(__name__)


def _str_or_none(string):
    """ Return a string of None if string is empty. """
    if string is None or string == "":
        return None
    return string


def _lookup(cls):
    """ Lookup module.class. """
    if isinstance(cls, str):
        module_name, _, name = cls.rpartition(".")
        module = importlib.import_module(module_name)
        try:
            cls = getattr(module, name)
        except AttributeError:
            raise AttributeError("%s reference cannot be found" % cls)
    return cls


class DataStore(base.DataStore):
    """ LDAP Account and group datastore. """

    def __init__(self, config):
        super(DataStore, self).__init__(config)
        self._using = config['LDAP']
        self._account = _lookup(config['ACCOUNT'])
        self._group = _lookup(config['GROUP'])
        self._primary_group = config.get(
            'PRIMARY_GROUP', 'institute')
        self._default_primary_group = config.get(
            'DEFAULT_PRIMARY_GROUP', 'dummy')
        self._home_directory = config.get(
            'HOME_DIRECTORY', "/home/%(uid)s")
        self._locked_shell = config.get(
            'LOCKED_SHELL', "/usr/local/sbin/locked")
        self._settings = config

    def _accounts(self):
        """ Return accounts query. """
        return self._account.objects.using(
            using=self._using, settings=self._settings)

    def _groups(self):
        """ Return groups query. """
        return self._group.objects.using(
            using=self._using, settings=self._settings)

    def _create_account(self, **kwargs):
        """ Create a new account. """
        return self._account(
            using=self._using, settings=self._settings, **kwargs)

    def _create_group(self, **kwargs):
        """ Create a new group. """
        return self._group(
            using=self._using, settings=self._settings, **kwargs)

    def save_account(self, account):
        """ Account was saved. """
        person = account.person
        if self._primary_group == 'institute':
            lgroup = self._groups().get(cn=person.institute.group.name)
        elif self._primary_group == 'default_project':
            if account.default_project is None:
                lgroup = self._groups().get(cn=self._default_primary_group)
            else:
                lgroup = self._groups().get(
                    cn=account.default_project.group.name)
        else:
            raise RuntimeError("Unknown value of PRIMARY_GROUP.")

        if account.default_project is None:
            default_project = "none"
        else:
            default_project = account.default_project.pid

        try:
            luser = self._accounts().get(uid=account.username)
            luser.gidNumber = lgroup.gidNumber
            luser.givenName = person.first_name
            luser.sn = person.last_name
            luser.fullName = person.full_name
            luser.telephoneNumber = _str_or_none(person.telephone)
            luser.mail = _str_or_none(person.email)
            luser.title = _str_or_none(person.title)
            luser.o = person.institute.name
            luser.gidNumber = lgroup.gidNumber
            luser.homeDirectory = self._home_directory % {
                'default_project': default_project,
                'uid': account.username}
            if account.is_locked():
                luser.loginShell = self._locked_shell
                luser.lock()
            else:
                luser.loginShell = account.shell
                luser.unlock()
            luser.save()
        except self._account.DoesNotExist:
            luser = self._create_account()
            luser.uid = account.username
            luser.givenName = person.first_name
            luser.sn = person.last_name
            luser.fullName = person.full_name
            luser.telephoneNumber = _str_or_none(person.telephone)
            luser.mail = _str_or_none(person.email)
            luser.title = _str_or_none(person.title)
            luser.o = person.institute.name
            luser.gidNumber = lgroup.gidNumber
            luser.homeDirectory = self._home_directory % {
                'default_project': default_project,
                'uid': account.username}
            if account.is_locked():
                luser.loginShell = self._locked_shell
                luser.lock()
            else:
                luser.loginShell = account.shell
                luser.unlock()
            luser.save()

            # add all groups
            for group in account.person.groups.all():
                self.add_account_to_group(account, group)

    def delete_account(self, account):
        """ Account was deleted. """
        try:
            luser = self._accounts().get(uid=account.username)
            luser.secondary_groups.clear()
            luser.delete()
        except self._account.DoesNotExist:
            # it doesn't matter if it doesn't exist
            pass

    def set_account_password(self, account, raw_password):
        """ Account's password was changed. """
        luser = self._accounts().get(uid=account.username)
        luser.change_password(raw_password)
        luser.save()

    def set_account_username(self, account, old_username, new_username):
        """ Account's username was changed. """
        luser = self._accounts().get(uid=old_username)
        luser.rename(uid=new_username)

    def add_account_to_group(self, account, group):
        """ Add account to group. """
        lgroup = self._groups().get(cn=group.name)
        person = self._accounts().get(uid=account.username)
        lgroup.secondary_accounts.add(person)

    def remove_account_from_group(self, account, group):
        """ Remove account from group. """
        lgroup = self._groups().get(cn=group.name)
        person = self._accounts().get(uid=account.username)
        lgroup.secondary_accounts.remove(person)

    def get_account_details(self, account):
        """ Get the account details. """
        result = {}
        try:
            luser = self._accounts().get(uid=account.username)
        except self._account.DoesNotExist:
            return result
        for i, j in luser.get_fields():
            if i != 'userPassword' and j is not None:
                result[i] = j
        result['dn'] = luser.dn
        group = luser.primary_group.get_obj()
        if group is not None:
            result['primary_group'] = group.dn
        result['secondary_groups'] = [
            g.dn for g in luser.secondary_groups.all()]
        return result

    def account_exists(self, username):
        """ Does the account exist? """
        try:
            self._accounts().get(uid=username)
            return True
        except self._account.DoesNotExist:
            return False

    def save_group(self, group):
        """ Group was saved. """
        # if group already exists, take over existing group rather then error.
        try:
            lgroup = self._groups().get(cn=group.name)
            lgroup.description = group.description
            lgroup.save()
        except self._group.DoesNotExist:
            lgroup = self._create_group()
            lgroup.cn = group.name
            lgroup.description = group.description
            lgroup.save()

    def delete_group(self, group):
        """ Group was deleted. """
        try:
            lgroup = self._groups().get(cn=group.name)
            lgroup.delete()
        except self._group.DoesNotExist:
            # it doesn't matter if it doesn't exist
            pass

    def set_group_name(self, group, old_name, new_name):
        """ Group was renamed. """
        lgroup = self._groups().get(cn=old_name)
        lgroup.rename(cn=new_name)

    def get_group_details(self, group):
        """ Get the group details. """
        result = {}
        try:
            lgroup = self._groups().get(cn=group.name)
        except self._group.DoesNotExist:
            return result
        for i, j in lgroup.get_fields():
            if j is not None:
                result[i] = j
        result['dn'] = lgroup.dn
        result['primary_accounts'] = [
            a.dn for a in lgroup.primary_accounts.all()]
        result['secondary_accounts'] = [
            a.dn for a in lgroup.secondary_accounts.all()]
        return result


trace.attach(trace.trace(logger), DataStore)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

""" This file shows the project application views using a state machine. """

from django.shortcuts import render
from django.http import HttpResponseBadRequest

from karaage.common.decorators import login_required
from karaage.people.models import Person

from karaage.plugins.kgapplications.views import base, states

from .forms import ApproveSoftwareForm
from .models import SoftwareApplication


class StateIntroduction(base.State):
    """ Invitation has been sent to applicant. """
    name = "Read introduction"

    def view(self, request, application, label, roles, actions):
        """ Django view method. """
        if label is None and \
                'is_applicant' in roles and 'is_admin' not in roles:
            for action in actions:
                if action in request.POST:
                    return action
            link, is_secret = base.get_email_link(application)
            return render(
                template_name='kgapplications/software_introduction.html',
                context={
                    'actions': actions,
                    'application': application,
                    'roles': roles,
                    'link': link,
                    'is_secret': is_secret
                },
                request=request)
        return super(StateIntroduction, self).view(
            request, application, label, roles, actions)


class StateWaitingForAdmin(states.StateWaitingForApproval):
    """ We need the administrator to provide approval. """
    name = "Waiting for administrator"
    authorised_text = "an administrator"
    authorised_role = "administrator"

    def get_authorised_persons(self, application):
        return Person.objects.filter(is_admin=True)

    def check_can_approve(self, request, application, roles):
        """ Check the person's authorization. """
        return 'is_admin' in roles

    def get_approve_form(self, request, application, roles):
        return ApproveSoftwareForm


def get_application_state_machine():
    """ Get the default state machine for applications. """
    open_transition = states.TransitionOpen(on_success='O')

    state_machine = base.StateMachine()
    state_machine.add_state(
        StateIntroduction(), 'O',
        {'cancel': 'R',
            'submit': states.TransitionSubmit(on_success='K', on_error="R")})
    state_machine.add_state(
        StateWaitingForAdmin(), 'K',
        {'cancel': 'R',
            'approve': states.TransitionApprove(
                on_password_needed='R', on_password_ok='C', on_error="R")})
    state_machine.add_state(
        states.StateCompleted(), 'C',
        {})
    state_machine.add_state(
        states.StateDeclined(), 'R',
        {'reopen': open_transition, })
    state_machine.set_first_state(open_transition)
    return state_machine


def register():
    base.setup_application_type(
        SoftwareApplication, get_application_state_machine())


@login_required
def new_application(request, software_license):
    # Called automatically by hook.
    assert software_license is not None

    if request.method != 'POST':
        return HttpResponseBadRequest("<h1>Bad Request</h1>")

    application = SoftwareApplication()
    application.applicant = request.user
    application.software_license = software_license
    application.save()

    state_machine = get_application_state_machine()
    response = state_machine.start(request, application, {})
    return response



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django.conf.urls import url, include
from django.conf import settings

from karaage.people.views import groups

urlpatterns = [
    url(r'^$', groups.group_list, name='kg_group_list'),
    url(r'^add/$', groups.add_group, name='kg_group_add'),
    url(r'^detail/(?P<group_name>%s)/' % settings.GROUP_VALIDATION_RE,
        include('karaage.people.urls.group_detail')),
]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django import forms
from django.conf import settings
from django.template import Context, Template

from karaage.people.models import Person
from karaage.projects.models import Project


EMAIL_GROUPS = (
    ('leaders', 'All Project Leaders (active projects only)'),
    ('users', 'All Active Users'),
    ('cluster_users', 'All Users with cluster accounts'),
)


class EmailForm(forms.Form):
    subject = forms.CharField(widget=forms.TextInput(attrs={'size': 60}))
    body = forms.CharField(widget=forms.Textarea(
        attrs={'class': 'vLargeTextField', 'rows': 10, 'cols': 40}))

    def get_data(self):
        return self.cleaned_data['subject'], self.cleaned_data['body']


class BulkEmailForm(EmailForm):
    group = forms.ChoiceField(choices=EMAIL_GROUPS)

    def get_emails(self):
        group = self.cleaned_data['group']
        subject, body = self.get_data()

        email_list = []

        subject_t = Template(subject)
        body_t = Template(body)
        emails = []

        if group == 'leaders':
            for p in Project.active.all():
                for leader in p.leaders.all():
                    ctx = Context({
                        'receiver': leader,
                        'project': p,
                    })
                    subject = subject_t.render(ctx)
                    body = body_t.render(ctx)
                    emails.append(
                        (subject, body, settings.ACCOUNTS_EMAIL,
                            [leader.email])
                    )
            return emails

        elif group == 'users':
            person_list = Person.active.all()

        elif group == 'cluster_users':
            person_list = Person.active.filter(account__isnull=False)

        else:
            person_list = None

        if person_list:
            person_list = person_list.filter(
                is_systemuser=False, login_enabled=True)

            for person in person_list:
                if person.email not in email_list:
                    ctx = Context({
                        'receiver': person,
                    })
                    subject = subject_t.render(ctx)
                    body = body_t.render(ctx)
                    emails.append(
                        (subject, body, settings.ACCOUNTS_EMAIL,
                            [person.email])
                    )
                    email_list.append(person.email)

            return emails

        return []



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django
if django.VERSION < (1, 7):
    parent = object
else:
    from django.apps import AppConfig
    parent = AppConfig


class BasePlugin(parent):
    name = None
    django_apps = ()
    depends = ()
    xmlrpc_methods = ()
    settings = {}
    template_context_processors = ()



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

""" This file implements a state machine for the views. """

from django.shortcuts import render
from django.core.urlresolvers import reverse
from django.conf import settings
from django.http import HttpResponseRedirect, HttpResponseForbidden
from django.http import HttpResponseBadRequest, HttpResponse, Http404
from django.contrib import messages

from karaage.common import log
import karaage.common as common

from ..models import Application


def get_url(request, application, roles, label=None):
    """ Retrieve a link that will work for the current user. """
    args = []
    if label is not None:
        args.append(label)

    # don't use secret_token unless we have to
    if 'is_admin' in roles:
        # Administrators can access anything without secrets
        require_secret = False
    elif 'is_applicant' not in roles:
        # we never give secrets to anybody but the applicant
        require_secret = False
    elif not request.user.is_authenticated():
        # If applicant is not logged in, we redirect them to secret URL
        require_secret = True
    elif request.user != application.applicant:
        # If logged in as different person, we redirect them to secret
        # URL. This could happen if the application was open with a different
        # email address, and the applicant is logged in when accessing it.
        require_secret = True
    else:
        # otherwise redirect them to URL that requires correct login.
        require_secret = False

    # return required url
    if not require_secret:
        url = reverse(
            'kg_application_detail',
            args=[application.pk, application.state] + args)
    else:
        url = reverse(
            'kg_application_unauthenticated',
            args=[application.secret_token, application.state] + args)
    return url


def get_admin_email_link(application):
    """ Retrieve a link that can be emailed to the administrator. """
    url = '%s/applications/%d/' % (settings.ADMIN_BASE_URL, application.pk)
    is_secret = False
    return url, is_secret


def get_registration_email_link(application):
    """ Retrieve a link that can be emailed to the logged other users. """
    url = '%s/applications/%d/' % (
        settings.REGISTRATION_BASE_URL, application.pk)
    is_secret = False
    return url, is_secret


def get_email_link(application):
    """ Retrieve a link that can be emailed to the applicant. """
    # don't use secret_token unless we have to
    if (application.content_type.model == 'person' and
            application.applicant.has_usable_password()):
        url = '%s/applications/%d/' % (
            settings.REGISTRATION_BASE_URL, application.pk)
        is_secret = False
    else:
        url = '%s/applications/%s/' % (
            settings.REGISTRATION_BASE_URL, application.secret_token)
        is_secret = True
    return url, is_secret


class StateMachine(object):
    """ State machine, for processing states. """

    ##################
    # PUBLIC METHODS #
    ##################

    def __init__(self):
        self._first_state = None
        self._states = {}
        super(StateMachine, self).__init__()

    def add_state(self, state, state_id, actions):
        """ Add a state to the list. The first state added becomes the initial
        state. """
        self._states[state_id] = state, actions

    def set_first_state(self, state_id):
        self._first_state = state_id

    def get_state(self, application):
        if application.state not in self._states:
            raise RuntimeError("Invalid state '%s'" % application.state)
        state, _ = self._states[application.state]
        return state

    def start(self, request, application, extra_roles=None):
        """ Continue the state machine at first state. """
        # Get the authentication of the current user
        roles = self._get_roles_for_request(request, application)
        if extra_roles is not None:
            roles.update(extra_roles)

        # Ensure current user is authenticated. If user isn't applicant,
        # leader, delegate or admin, they probably shouldn't be here.
        if 'is_authorised' not in roles:
            return HttpResponseForbidden('<h1>Access Denied</h1>')

        # Sanity check
        if self._first_state is None:
            raise RuntimeError("First state not set.")

        # Go to first state.
        return self._next(request, application, roles, self._first_state)

    def process(
            self, request, application,
            expected_state, label, extra_roles=None):
        """ Process the view request at the current state. """

        # Get the authentication of the current user
        roles = self._get_roles_for_request(request, application)
        if extra_roles is not None:
            roles.update(extra_roles)

        # Ensure current user is authenticated. If user isn't applicant,
        # leader, delegate or admin, they probably shouldn't be here.
        if 'is_authorised' not in roles:
            return HttpResponseForbidden('<h1>Access Denied</h1>')

        # If user didn't supply state on URL, redirect to full URL.
        if expected_state is None:
            url = get_url(request, application, roles, label)
            return HttpResponseRedirect(url)

        # Check that the current state is valid.
        if application.state not in self._states:
            raise RuntimeError("Invalid state '%s'" % application.state)

        # If state user expected is different to state we are in, warn user
        # and jump to expected state.
        if expected_state != application.state:
            # post data will be lost
            if request.method == "POST":
                messages.warning(
                    request,
                    "Discarding request and jumping to current state.")
            # note we discard the label, it probably isn't relevant for new
            # state
            url = get_url(request, application, roles)
            return HttpResponseRedirect(url)

        # Get the current state for this application
        state, actions = self._states[application.state]

        # Finally do something
        if request.method == "GET":
            # if method is GET, state does not ever change.
            response = state.view(
                request, application, label, roles, actions.keys())
            assert isinstance(response, HttpResponse)
            return response

        elif request.method == "POST":
            # if method is POST, it can return a HttpResponse or a string
            response = state.view(
                request, application, label, roles, actions.keys())
            if isinstance(response, HttpResponse):
                # If it returned a HttpResponse, state not changed, just
                # display
                return response
            else:
                # If it returned a string, lookit up in the actions for this
                # state
                if response not in actions:
                    raise RuntimeError(
                        "Invalid response '%s' from state '%s'"
                        % (response, state))
                next_state_key = actions[response]
                # Go to the next state
                return self._next(request, application, roles, next_state_key)

        else:
            # Shouldn't happen, user did something weird
            return HttpResponseBadRequest("<h1>Bad Request</h1>")

    ###################
    # PRIVATE METHODS #
    ###################
    @staticmethod
    def _get_roles_for_request(request, application):
        """ Check the authentication of the current user. """
        roles = application.get_roles_for_person(request.user)

        if common.is_admin(request):
            roles.add("is_admin")
            roles.add('is_authorised')

        return roles

    def _next(self, request, application, roles, state_key):
        """ Continue the state machine at given state. """
        # we only support state changes for POST requests
        if request.method == "POST":
            # If next state is a transition, process it
            while isinstance(state_key, Transition):
                state_key = state_key.get_next_state(
                    request, application, roles)

            # lookup next state
            if state_key not in self._states:
                raise RuntimeError("Invalid state '%s'" % state_key)
            state, _ = self._states[state_key]

            # enter that state
            state.enter_state(request, application)
            application.state = state_key
            application.save()

            # log details
            log.change(application.application_ptr, "state: %s" % state.name)

            # redirect to this new state
            url = get_url(request, application, roles)
            return HttpResponseRedirect(url)
        else:
            return HttpResponseBadRequest("<h1>Bad Request</h1>")


class State(object):
    """ A abstract class that is the base for all application states. """
    name = "Abstract State"

    def __init__(self):
        self.context = {}

    def enter_state(self, request, application):
        """ This is becoming the new current state. """
        pass

    def view(self, request, application, label, roles, actions):
        """ Django view method. We provide a default detail view for
        applications. """

        # We only provide a view for when no label provided
        if label is not None:
            return HttpResponseBadRequest("<h1>Bad Request</h1>")

        # only certain actions make sense for default view
        tmp_actions = []
        if 'reopen' in actions:
            tmp_actions.append("reopen")
        if 'cancel' in actions:
            tmp_actions.append("cancel")
        if 'duplicate' in actions:
            tmp_actions.append("duplicate")
        if 'archive' in actions and 'is_admin' in roles:
            tmp_actions.append("archive")
        actions = tmp_actions

        # process the request in default view
        if request.method == "GET":
            context = self.context
            context.update({
                'application': application,
                'actions': actions,
                'state': self.name,
                'roles': roles})
            return render(
                template_name='kgapplications/common_detail.html',
                context=context,
                request=request)
        elif request.method == "POST":
            for action in actions:
                if action in request.POST:
                    return action

        # we don't know how to handle this request.
        return HttpResponseBadRequest("<h1>Bad Request</h1>")


class Transition(object):
    """ A transition from one state to another. """

    def __init__(self):
        pass

    def get_next_state(self, request, application, roles):
        """ Retrieve the next state. """
        raise NotImplementedError()


def get_application(**kwargs):
    try:
        application = Application.objects.get(**kwargs)
    except Application.DoesNotExist:
        application = None

    try:
        if application is not None:
            application = application.get_object()
            return application
    except Application.DoesNotExist:
        raise RuntimeError("The application is currupt.")

    raise Http404("The application does not exist.")


_types = {}


def setup_application_type(application_type, state_machine):
    _types[application_type] = state_machine


def get_state_machine(application):
    application = application.get_object()
    return _types[type(application)]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from karaage.tests.integration import IntegrationTestCase
from karaage.tests.fixtures import AccountFactory


class OpenldapAccountTestCase(IntegrationTestCase):

    def setUp(self):
        super(OpenldapAccountTestCase, self).setUp()

    def test_kAccountMixin(self):
        account = AccountFactory()
        ldap = self._ldap_datastore._accounts()
        self._ldap_datastore.save_account(account)
        ldap_account = ldap.get(uid=account.username)
        self.assertEqual(
            ldap_account.uid,
            account.username
        )
